/*

QLTOOLS 

Read a QL floppy disk

(c)1992 by Giuseppe Zanetti

Giuseppe Zanetti
via Vergani, 11 - 35031 Abano Terme (Padova) ITALY
e-mail: beppe@sabrina.dei.unipd.it

*/

#define VERSION     "1.6,  Jan 31 1992"

/* Maximum allocation block (normally 3) */

#define MAXALB		6

/* Maximum number of sectors (norm. 1440) */

#define MAXSECT		2880

#include <stdio.h>

#ifdef THINK_C
#include "console.h"
#include <string.h>
#else
#include <string.h>
#endif

#ifdef MSDOS
#include <bios.h>
#include <dos.h>    /* for delay    */
#include <io.h>     /* for setmode  */
#include <fcntl.h>  /* for O_BINARY */
#define	RESET	0
#define LAST	1
#define READ	2
#define	WRITE	3
#define VERIFY	4
#define FORMAT	5
#endif

/* this line is neccessary since stdio.h in SUN 4.0 do not define it */

#ifndef SEEK_SET
#define SEEK_SET	0
#endif

FILE *f;
static unsigned char b[512*MAXALB];
static unsigned char b0[512*MAXALB];
static unsigned char ltp[18];
static unsigned char ptl[18];
static unsigned int bleod;
static unsigned int byeod;
static unsigned char *pdir;

static unsigned int convtable[MAXSECT];

#ifdef MSDOS
static unsigned int drive;
static unsigned int conv_side[MAXSECT];
static unsigned int conv_sector[MAXSECT];
static unsigned int conv_track[MAXSECT];
#endif


static int gsides,gtracks,gsectors,goffset,allocblock;

unsigned int wrd(wp)
unsigned char *wp;
{
    unsigned int r;

    r= ((unsigned int) *(wp))*256 + (unsigned int) *(wp+1);

    return r;
}

unsigned long int lwrd(p)
unsigned char *p;
{
    unsigned long int r,h,l;

    h=(unsigned long int) *(p)*256 + *(p+1);
    l=(unsigned long int) *(p+2)*256 + *(p+3);


    r = (unsigned long int) (h * 65536 + l);

    return(r);
}

void cat_file(fnum)
long int fnum;
{
	long int flen,file,blk;
	int i,s,c,start,end;
	char buffer[512*MAXALB];
	unsigned char *base;
	int ok;

#ifdef MSDOS
    setmode(fileno(stdout),O_BINARY);
#endif

    flen=lwrd(pdir+fnum*64);    /* with the header */

	for (s=0;s<=flen/(512*allocblock);s++)
	{
		ok=0;

        	for (i=0;i<(wrd(b0+28)*wrd(b0+30))/wrd(b0+32);i++)
        	{
			base=b0+0x4C+20+i*allocblock;
			file=(int) (*(base) * 0x10) + (int) ((*(base+1) >> 4) & 0x0F) ;
			blk= (*(base+1) & 0x0F)*0x100 + *(base+2);
 
			if (file == fnum)
			{
				if (blk == s)
				{
					read_block(buffer,i);

					start=0; if (s==0) start=64;
					end=512*allocblock; if (s==(flen/(512*allocblock))) end=flen % (512*allocblock);

                    for(c=start;c<end;c++) printf("%c",*(buffer+c));

					ok=1;
				}
			}
		}

		if (!ok)
		{
			fprintf(stderr,"\n\nBlock #%i of file #%i not found\n\n",s,fnum);

			if (s==0) fprintf(stderr,"I think this file was erased\n\n");

#ifdef MSDOS
    setmode(fileno(stdout),O_BINARY);
#endif

			exit(1);
		}
	}

#ifdef MSDOS
    setmode(fileno(stdout),O_BINARY);
#endif

}

void usage(error)
char *error;
{
	fprintf(stderr,"error %s\n\n",error);

#ifdef MSDOS
    fprintf(stderr,"Usage: qltools [a:|b:] -[options] [filename]\n\n");
#else
    fprintf(stderr,"Usage: qltools diskimage -[options] [filename]\n\n");
#endif

	fprintf(stderr,"options:\n\n");
	fprintf(stderr,"-d      list directory\n");
	fprintf(stderr,"-s      list short directory\n");
	fprintf(stderr,"-i      list info\n");
	fprintf(stderr,"-m      list disk map\n");
	fprintf(stderr,"-c      list conversion table of sectors\n");
	fprintf(stderr,"-nFN    output file number FN in the directory listing\n\n");

#ifdef MSDOS
    fprintf(stderr,"QLTOOLS for MS-DOS (version ");
    fprintf(stderr,VERSION);
    fprintf(stderr,")\n\n");
#else
    fprintf(stderr,"diskimage is either a file with the image of a QL format disk\n");
    fprintf(stderr,"or a Unix device with a QL disk inserted in it (/dev/fd...)\n\n");
    fprintf(stderr,"QLTOOLS for UNIX (version: ");
    fprintf(stderr,VERSION);
    fprintf(stderr,")\n\n");
#endif

	fprintf(stderr,"Giuseppe Zanetti\n");
    fprintf(stderr,"via Vergani, 11 - 35031 Abano Terme (Padova) ITALY\n");
	fprintf(stderr,"e-mail: beppe@sabrina.dei.unipd.it\n");
	exit(1);
}


void print_info()
{
	int i;

	printf("Disk ID          : ");

	for (i=0;i<4;i++)
	{
		printf("%c",b0[i]);
	}
	printf("\n");

	printf("Disk Label       : ");

	for (i=0;i<10;i++)
	{
		printf("%c",b0[4+i]);
	}
	printf("\n");

	printf("sectors per track: %i\n",gsectors);
	printf("sectors per cyl. : %i\n",wrd(b0+28));
	printf("number of cylind.: %i\n",gtracks);
	printf("allocation block : %i\n",allocblock);
	printf("sector offset/cyl: %i\n",goffset);

	printf("free sectors     : %i\n",wrd(b0+20));
	printf("good sectors     : %i\n",wrd(b0+22));
	printf("total sectors    : %i\n",wrd(b0+24));

    printf("directory is     : %u sectors and %u bytes\n",bleod,byeod);

	printf("\nlogical-to-physical sector mapping table:\n\n");
	for (i=0;i<18;i++) printf("%u ",ltp[i]);
	printf("\n");

	printf("\nphysical-to-logical sector mapping table:\n\n");
	for (i=0;i<18;i++) printf("%u ",ptl[i]);
	printf("\n");
}

void print_dir(shortflag)
int shortflag;
{
	int d,i;
	char c;

	if (!shortflag)
	{
		for (i=0;i<10;i++)
	        {
	                printf("%c",b0[4+i]);
	        }
	        printf("\n");

		printf("%i/%i sectors.\n\n",wrd(b0+20),wrd(b0+22));
	}

	for (d=64;d <512*bleod+byeod;d+=64)
	{
		if(lwrd(pdir+d)==0)
			continue;

		if (!shortflag)
		{
			printf("%3i : ",d/64);
		}

		for (i=0;i<wrd(pdir+d+14);i++)
		{
			c=*(pdir+d+16+i);

			printf("%c",c);
		}

		if (!shortflag)
		{
			for(i=0;i<38-wrd(pdir+d+14);i++) printf(" ");

			switch(*(pdir+d+5))
			{
				case 0: printf("          "); break;
				case 1: printf("Exec      "); break;
				case 2: printf("Rel       "); break;
				case 255:printf("->        "); break;
				default:printf("(type %3d)",*(pdir+d+5)); break;
			}

		        printf("    %ld",lwrd(pdir+d) - 64);
		}

		printf("\n");
	}
}

void make_convtable(verbose)
int verbose;
{
	int i,si,tr,se,ls,ps,uxs;

	if (verbose)
	{
		printf("\nCONVERSION TABLE\n\n");
		printf("logic\ttrack\tside\tsector\tunix_dev\n\n");
	}

	for(i=0;i<gtracks*gsectors*gsides;i++)
	{
		tr=i / (gsectors*gsides);
		ls=i % (gsectors*gsides);

		ps = ltp[ls];

		si=(ps & 0x80) != 0;

		ps &= 0x7F;

		ps += goffset*tr;
		se = ps % gsectors;

		uxs = tr*gsectors*gsides+gsectors*si+se;

		convtable[i] = uxs;

#ifdef MSDOS
conv_side[i]=si;
conv_sector[i]=se;
conv_track[i]=tr;
#endif

		if (verbose)
		{
			printf("%i\t%i\t%i\t%i\t%i\n",i,tr,si,se,uxs);
		}
	}
}

int read_block(p,num)
char *p;
int num;
{
	int i;
	int r=0;

	for(i=0;i<allocblock;i++)
	{

#ifdef MSDOS
		/* biosdisk(RESET, drive , 0, 0, 0, 0, p+512*i); */
        biosdisk(READ, drive ,conv_side[num*allocblock+i],conv_track[num*allocblock+i],1+conv_sector[num*allocblock+i],1,p+512*i);
#else
		fseek(f,(512*convtable[num*allocblock+i]),SEEK_SET);
		r += fread(p+512*i,512,1,f);
#endif

	}

	return(r);
}

int read_block0(argconv)
int argconv;
{
	int i;

#ifdef MSDOS

    int status;

    /* reset the disk */

    status = biosdisk(READ, drive, 0, 10, 1, 1, b0);

    /* door change signal ? */

    if (status == 0x06) status = biosdisk(READ, drive, 0, 0, 1, 1, b0);

    status = biosdisk(RESET, drive , 0, 0, 0, 0, b0);

    if (status != 0) fprintf(stderr,"Disk not ready (continuing...)\n");

    /* read block 0 */

    biosdisk(READ, drive ,0,0,1,1,b0);
#else
	fseek(f,0,SEEK_SET);
	fread(b0,512,1,f);
#endif

	/* is this a QL disk ? */

	if ((*(b0) != 'Q') || (*(b0+1) != 'L'))
	{
		fprintf(stderr,"\nNot a QL disk !!!\n\n");
		exit(1);
	}

	/* read ptl and ptl tables */

	for (i=0;i<18;i++)
	{
		ltp[i] = *(b0+40+i);
		ptl[i] = *(b0+58+i);
	}

	gsides=2;
	gtracks=wrd(b0+30);
	gsectors=wrd(b0+26);
	goffset=wrd(b0+38);
	bleod=wrd(b0+34);
	byeod=wrd(b0+36);
	allocblock=wrd(b0+32);

	make_convtable(argconv);

	read_block(b0,0);

	return(0);
}

long int match_file(fname)
char *fname;
{
	int d,i,match,len;
	long int r=0L;
	char c;

	len=strlen(fname);

	for (d=64;d <512*bleod+byeod;d+=64)
	{
		match=1;

		if (wrd(pdir+d+14) == len)
		{
			for (i=0;i<len;i++)
			{
				c=*(pdir+d+16+i);

				if (c != fname[i]) match=0;
			}

			if (match)
			{
				r=d/64;
				break;
			}
		}
	}

	return(r);
}

main(argc,argv)
int argc;
char * argv[];
{
	int i;
	unsigned long int finfofile, finfoblk;
	unsigned char *base;

	int argdir,arginfo,argfnum,argmap,argconv,argshort;
	char argfname[255];

#ifdef THINK_C
	argc=ccommand(&argv);
#endif

	argdir=arginfo=argfnum=argmap=argconv=argshort=0;
	strcpy(argfname,"");

	if (argc < 2)
	{
		usage("too few parameters");
	}

	for(i=2;i<argc;i++)
	{
        if ( (argv[i][0] == '-') || (argv[i][0] == '/') )
		{
			switch(argv[i][1])
			{
				case 'd': argdir=1; break;
				case 's': argdir=1; argshort=1; break;
				case 'i': arginfo=1; break;
				case 'm': argmap=1; break;
				case 'c': argconv=1; break;
				case 'n': argfnum=atol(argv[i]+2); break;
				default: usage("bad option"); break;
			}
		}
		else
		{
			strcpy(argfname,argv[i]);
		}
	}

#ifdef MSDOS
	if (argv[1][1] != ':')
	{
		fprintf(stderr,"Bad drive %s\n",argv[1]);
		usage("Bad drive: use a: or b:");
	}

	switch(argv[1][0])
	{
		case 'a':
		case 'A': drive=0; break;

		case 'b':
		case 'B': drive=1; break;

		default: usage("Bad drive: use a: or b:"); break;
	}
#else
	f=fopen(argv[1],"rb");

	if (f==NULL)
	{
		usage("bad image file or device");
	}
#endif

	read_block0(argconv);

	pdir=(unsigned char *) malloc(512*allocblock*(bleod+1));

	/* read the directory map */

	if (argmap)
	{
		printf("\nblock\tfile\tpos\n\n");
	}

	for (i=0;i<(wrd(b0+28)*wrd(b0+30))/wrd(b0+32);i++)
	{
		base=b0+0x4C+20+i*allocblock;

		finfofile= (unsigned long int) (*(base) * 0x10) + (unsigned long int) ((*(base+1) >> 4) & 0x0F) ;
		finfoblk= (*(base+1) & 0x0F)*0x100 + *(base+2);


		if (finfofile == 0x00)
		{
            read_block(pdir+512*allocblock*finfoblk,i);
		}

		if (argmap)
		{
			printf("%d\t%ld\t%ld\t",i,finfofile,finfoblk);

			if ((*(base) & 0xFD) == 0xFD)
			{
				printf("erased ");
			}

			switch(finfofile)
			{
				case 0x000: printf("directory %i\n",finfoblk); break;
				case 0xF80: printf("map\n"); break;
				case 0xFDF: printf("unused\n"); break;
				case 0xFEF: printf("bad\n"); break;
				case 0xFFF: printf("not existent\n"); break;
				default: printf("\n"); break;
			}

		}
	}

	if (arginfo) print_info();
	if (argdir) print_dir(argshort);
	if (argfnum != 0) cat_file((long int)argfnum);
	if (strcmp(argfname,"") != 0)
	{
		argfnum=match_file(argfname);

		if (argfnum == 0)
		{
			fprintf(stderr,"file not found\n");
			exit(2);
		}

	cat_file((long int)argfnum);
	}

#ifdef MSDOS
#else
	fclose(f);
#endif

	return (0);
}

